/*

    textures: a utility to merge new textures with textures from DOOM.WAD
    Copyright (C) 1994  Robert H. Forsman Jr.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "wad.h"

#define TEXTURE_SIZE(npatches)	(sizeof(struct dw_one_texture) + (npatches-1)*sizeof(struct dw_patchdesc))

/* this structure stores all the information necessary to interpret a
   texture resource.  The pnames are clumped in with the texture data
   because the texture data contains indexes into the pnames and we
   load several different pnames which go with different texture data.
   */

struct texture_gob {
    struct dw_texture_list	*tdata;		/* TEXTURE resource */
    INT32	tlen;
    /* the pndata can be shared by several textures */
    struct dw_pnames	*pndata;	/* PNAMES resource */
	 INT32	pnlen;
};

/* all the texture/pname gobs we've loaded in */
struct texture_gob	*allgobs=0;
int	numgobs=0;

/* we're about to add another gob.  make sure there's space and give
   us a pointer */
struct texture_gob *extra_gob(void)
{
    if (allgobs) {
	allgobs = (struct texture_gob*)realloc(allgobs,
					       sizeof(*allgobs)*(numgobs+1));
	return &allgobs[numgobs++];
    } else {
	allgobs = (struct texture_gob*)malloc(sizeof(*allgobs)*(numgobs=1));
	return &allgobs[0];
    }
}

/* This is the entire reason this program exists.  These resources are
   constructed and mangled together from the input files.  The result
   is a single texture resource and a single pnames resource that can
   be written to a final output WAD file */
struct dw_texture_list	*output_texture;
int	otlen;
struct dw_pnames	*output_pnames;
int	oplen;

void compile_textures(void)
{
    char	(*tnames)[8];
    int		*offsets;
    int		ntnames=0;
    char	(*pnames)[8];
    int		npnames=0;
    int		*idxcache;

    char	*data;
    int/*damn, is this big enough?*/	datalen, datasize;

    int	i,j,k;

    datalen = 0;
    data = malloc(datasize=4096); /* don't shrink this */

    /* just to make sure realloc doesn't puke */
    offsets = malloc(1);
    pnames = malloc(1);
    tnames = malloc(1);

    for (i=numgobs-1; i>=0; i--) {
	idxcache = malloc(sizeof(*idxcache) * (size_t)allgobs[i].pndata->count);

	for (j=0; j<allgobs[i].pndata->count; j++) {
	    for (k=0; k<npnames; k++)
		if (0==strncasecmp(pnames[k],
				   allgobs[i].pndata->names[j], 8))
		    break;
	    if (k<npnames)
		idxcache[j] = k;
	    else {
		pnames = realloc(pnames, sizeof(*pnames)*(++npnames));
		idxcache[j] = k;
		strncpy(pnames[k], allgobs[i].pndata->names[j], 8);
	    }
	}

	if (allgobs[i].tdata)
	    for (j=0; j<allgobs[i].tdata->ntextures; j++) {
		struct dw_one_texture	*curr;
		if (4+j*4 >= allgobs[i].tlen ||
		    Itohl(allgobs[i].tdata->offsets[j]) >= allgobs[i].tlen ||
		    Itohl(allgobs[i].tdata->offsets[j]) + sizeof(struct dw_one_texture) > allgobs[i].tlen) {
		    fprintf(stderr, "corrupt texture list in gob #%d\n", i);
		    break;
		}
		curr = (struct dw_one_texture*)(((char*)allgobs[i].tdata)
						+ (int)Itohl(allgobs[i].tdata->
							offsets[j]));
		if (Itohl(allgobs[i].tdata->offsets[j]) +
		    sizeof(struct dw_one_texture) +
		    (curr->npatches-1)*sizeof(struct dw_patchdesc)
			 > allgobs[i].tlen) {
		    fprintf(stderr, "corrupt texture list in gob #%d\n", i);
		    break;
		}
		for (k=0; k<ntnames; k++) {
		    if (0==strncasecmp(tnames[k], curr->name, 8))
			break;
		}
		if (k<ntnames) {
		    fprintf(stderr, "ignoring duplicate texture %s\n",
			    curr->name);
		    continue;
		}
		tnames = realloc(tnames, sizeof(*tnames)*(++ntnames));
		offsets = realloc(offsets, sizeof(*offsets)*ntnames);
		offsets[ntnames-1] = datalen;
		strncpy(tnames[k], curr->name, 8);

		/* this following process is destructive! */
		for (k=0; k<curr->npatches; k++) {
		    curr->patches[k].pnames_idx
			= htoIs(idxcache[Itohs(curr->patches[k].pnames_idx)]);
		}

		while (datalen+TEXTURE_SIZE(curr->npatches) >= datasize)
		    data = realloc(data, datasize*=2);

		memcpy(data+datalen, curr, TEXTURE_SIZE(curr->npatches));

		datalen += TEXTURE_SIZE(curr->npatches);
	    }

	free(idxcache);
    }

    output_texture = malloc(otlen = 4+4*ntnames+ datalen);
    output_texture->ntextures = htoIl(ntnames);
    for (i=0; i<ntnames; i++)
	output_texture->offsets[i] = htoIl(4+4*ntnames+offsets[i]);
    free(offsets);
    memcpy(&output_texture->offsets[ntnames], data, datalen);
    free(data);

    output_pnames = malloc(oplen = 4+8*npnames);
    output_pnames->count = htoIl(npnames);
    memcpy(output_pnames->names, pnames, 8*npnames);
}

/**********************************************************************/

/* this procedure prints the compiled texures into an ASCII form that
   is self-contained (it doesn't need a PNAMES).  The format output by
   this procedure is compatible with this programs -at format */
void pretty_print(FILE *fp)
{
    struct dw_texture_list	*tl;
    int	i,j;

    tl = output_texture;

    for (i=0; i< Itohl(tl->ntextures); i++) {
	struct dw_one_texture	*curr;
	curr = (struct dw_one_texture *)((char*)(tl) + (int)Itohl(tl->offsets[i]));
	fprintf(fp, "%-8s : %dx%d\n", curr->name,
		Itohs(curr->width), Itohs(curr->height));
	for (j=0; j<Itohs(curr->npatches); j++) {
	    struct dw_patchdesc	*patch = &curr->patches[j];
	    char	buf[9];
	    upcopy8(buf, output_pnames->names[Itohs(patch->pnames_idx)]);
	    buf[8] = 0;
	    fprintf(fp, "\t%-8s %+d%+d %d %d\n", buf,
		    Itohs(patch->xoff), Itohs(patch->yoff),
		    Itohs(patch->stepdir), Itohs(patch->colormap));
	}
    }
}

/* This procedure reads a list of patch names from the specified file.
   It constructs a PNAMES resource from them */
struct dw_pnames *read_ascii_pnames(char *fname, INT32 *len)
{
    FILE	*fp;
    struct dw_pnames	*rval;
    int	size;
	 fp = fopen_b(fname, 0, 0);
	 if (fp==0) {
	*len = 0;
	return 0;
    }

    size = 4096;
	 rval = malloc(size);
    rval->count = 0;
    while (!feof(fp)) {
	int	j, ch;

	if (size <4+(rval->count+1)*8)
	    rval = realloc(rval, size*=2);
	for (j=0; j<8; j++) {
	    ch = getc(fp);
	    if (ch==EOF || ch=='\n')
		break;
		 rval->names[(int)rval->count][j] = islower(ch) ? toupper(ch) : ch;
	}
	if (j<8) {
	    for (; j<8; j++)
		rval->names[(int)rval->count][j] = 0;
	} else {
	    int	first =1;
	    while (1) {
		ch = getc(fp);
		if (ch==EOF||ch=='\n')
		    break;
		if (first) {
		    fprintf(stderr, "pname on line %ld of file %s too long\n",
			    rval->count+1, fname);
		    first = 0;
		}
	    }
	}
	rval->count++;
    }

    fclose_b(fp);

	 rval = realloc(rval, (size_t)(*len = 4 + rval->count*8));
    rval->count = htoIl(rval->count);

    return rval;
}

/* This is used by make_gob_from_textfile.  Since the ascii texture
   file format contains patch NAMES rather than indices, when we read
   it into core memory we must convert these names into indices. */
INT32 maybe_add_pname(char *name, struct texture_gob *gob)
{
  int	i;

  if (gob->pndata==0) {
	 gob->pndata = malloc( (size_t)gob->pnlen = 1024);
    gob->pndata->count = 0;
  }
    
  for (i=0; i<gob->pndata->count; i++) {
    if (0==strncasecmp(name, gob->pndata->names[i], 8))
      break;
  }

  if (i<gob->pndata->count)
    return i;

  if (gob->pnlen < (gob->pndata->count+1)*8 + 4)
	 gob->pndata = realloc(gob->pndata, (size_t)gob->pnlen*=2);

  upcopy8(gob->pndata->names[i], name);
  return gob->pndata->count++;
}

/* creates a TEXTURE and PNAMES resource from an ascii file.  The file
   has several "paragraph"s.  Each paragraph describes one texture.

   The first line of the pargraph has the texture name, a colon, the
   width, an 'x', then the height.

   The subsequent lines of the paragraph begin with a TAB character
   and contain the patch name, two integers for the patch X and Y
   offset, the stepdir (always 1) and the colormap (always 0)

   The first line that doesn't start with a TAB must belong to the
   next paragraph.

   No extraneous blank lines are allowed, and there are no such things
   as comments.  If this is a serious problem for you, drop me a line.
   I'll see what I can do.

*/

void make_gob_from_textfile(char *fname, struct texture_gob *gob)
{
    FILE	*fp;
    char	*data;
    int		datalen, datasize;
    int		*offsets, noffsets;
    int		numtextures;
    int		linen;
    char	buf[1024];

    gob->tdata=0;
    gob->tlen =0;
    gob->pndata=0;
    gob->pnlen=0;

    if (!(fp = fopen_b(fname, 0, 0)))
	return;

    datalen=0;
    data = malloc(datasize=4096);

#define	TEXTURE	((struct dw_one_texture*)(data+datalen))

    numtextures = 0;
    noffsets=16;
    offsets = malloc(sizeof(*offsets)*noffsets);
    linen=1;			/* jumpstart the input loop */
    if (0==fgets(buf, sizeof(buf), fp))
	return;
    while (!feof(fp)) {
	char	name[9];
	int	width,height;
	int	npatches;

	if (3!=sscanf(buf, "%8s : %dx%d", name, &width, &height)) {
	    fprintf(stderr, "error parsing ascii texture file %s on line %d\n",
		    fname, linen);
	    return;
	}

	if (datalen+sizeof(struct dw_one_texture) >= datasize)
	    data = realloc(data, datasize *=2);
	if (numtextures >= noffsets)
	    offsets = realloc(offsets, sizeof(*offsets)*(noffsets*=2));
	offsets[numtextures] = datalen;
	numtextures++;

	upcopy8(TEXTURE->name, name);
	TEXTURE->width  = htoIs(width);
	TEXTURE->height = htoIs(height);
	TEXTURE->zero1 = TEXTURE->zero2 = TEXTURE->zero3 = TEXTURE->zero4 = 0;
	for (npatches=0; ; npatches++) {
	    int	xoff, yoff, stepdir, colormap;
	    if (0==fgets(buf, sizeof(buf), fp)) {
		buf[0] = 0;
		break;
	    }
	    linen++;
	    if (buf[0]!='\t')
		break;
	    
	    if (5!=sscanf(buf, "\t%8s  %d%d %d %d", name,
			  &xoff, &yoff, &stepdir, &colormap)) {
		fprintf(stderr, "error parsing patch from ascii texture file %s on line %d\n",
			fname, linen);
		return;
	    }
	    if (datalen+TEXTURE_SIZE(npatches+1) >= datasize) {
		data = realloc(data, datasize *=2);
	    }
	    TEXTURE->patches[npatches].xoff = htoIs(xoff);
	    TEXTURE->patches[npatches].yoff = htoIs(yoff);
	    TEXTURE->patches[npatches].stepdir = htoIs(stepdir);
	    TEXTURE->patches[npatches].colormap = htoIs(colormap);
	    TEXTURE->patches[npatches].pnames_idx =
		htoIs(maybe_add_pname(name, gob));
	}
	TEXTURE->npatches = htoIs(npatches);
	datalen += TEXTURE_SIZE(npatches);
    }

	 gob->tdata = malloc( (size_t)gob->tlen = 4 + 4*numtextures + datalen);
    gob->tdata->ntextures = htoIl(numtextures);
    {
	int i;
	for (i=0; i<numtextures; i++)
	    gob->tdata->offsets[i] = htoIl(offsets[i]+4+4*numtextures);
    }
    memcpy( 4 + 4*numtextures + (char*)gob->tdata, data, datalen);

    /* squish down the pnames */
	 gob->pnlen = 4+8*gob->pndata->count;
	 gob->pndata = realloc(gob->pndata, (size_t)gob->pnlen);

    free(offsets);
    free(data);

    fclose_b(fp);
}

/**********************************************************************/

struct wadcache {
	 INT32	nentries;
    union {
	/* this structure caches the directory of a WAD file.  We do
	   not load the entire file into core because DOS is pretty
	   darn short on memory.  We keep the FILE pointer so we can
	   later seek through it reading in resources. */
	struct {
	    char	*passthru;	/* should these be passed through to
					   the output wad? */
	    struct dw_direntry *entries;
	    FILE	*fp;
	} realwad;
	/* this structure holds one single raw resource. */
	struct {
	    char	*data;
		 INT32		len;
	    char	name[8];
	} oneres;
    }u;
};

/* a list of WAD files and individual resources */
struct wadcache *allwads=0;
int	nwads=0;


struct wadcache *new_wadcache(void)
{
    if (allwads) {
	allwads = realloc(allwads, sizeof(*allwads)*(nwads+1));
    } else {
	allwads = malloc(sizeof(*allwads));
    }
    return &allwads[nwads++];
}

/* This procedure loads an input wad, extracts the pnames and any
   texture resources that might be in it, and (if the user specifies a
   +) stores the directory so that the resources may be later copied
   from the file into an output wad file. */

/* argv[*argi] is the "-" or "+" string when we're done */
void slurp_iwad(char **argv, int *argi)
{
    FILE	*fp;
    char	magic[4];
    INT32	nentries;
    INT32	offset;
    struct wadcache	*curr;

	 if (!(fp = fopen_b(argv[*argi+1], 0, 1)))
	exit(1);

    if (1!=fread(magic, 4, 1, fp) ||
	1!=fread(&nentries, 4, 1, fp) ||
	1!=fread(&offset, 4, 1, fp)) {
	if (0!=memcmp("IWAD", magic, 4) &&
	    0!=memcmp("PWAD", magic, 4)) {
	    fprintf(stderr, "File %s not an IWAD or PWAD\n", argv[*argi+1]);
	    fclose_b(fp);
	    exit(1);		/* we're not eating *argi */
	}
    }
	 if (Itohs(nentries)<1) {
	return;			/* nothing interesting here */
    }

	 curr = new_wadcache();

	 curr->nentries = nentries = Itohl(nentries);
	 curr->u.realwad.entries = malloc(sizeof(*curr->u.realwad.entries)
					  *(size_t)nentries);
	 curr->u.realwad.passthru = malloc(sizeof(*curr->u.realwad.passthru)
						*(size_t)nentries);
    curr->u.realwad.fp = fp;

    fseek(fp, (long) Itohl(offset), SEEK_SET);

    if (nentries != fread(curr->u.realwad.entries,
			  sizeof(*curr->u.realwad.entries),
			  (size_t)nentries, fp)) {
	fprintf(stderr, "short read on WAD directory entries\n");
	exit(1);
    }

    if (0==strcmp(argv[2+*argi], "+")) {
	int	j;
	for (j=0; j<curr->nentries; j++)
	    curr->u.realwad.passthru[j] = 1;

	*argi += 2;

    } else {
	int	i;
	int	j;
	struct dw_pnames	*pndata;
	long	pnlen;

	for (j=0; j<curr->nentries; j++)
	    curr->u.realwad.passthru[j] = 1;

	for (j=0; j<curr->nentries; j++) {
	    if (0==strncasecmp(curr->u.realwad.entries[j].name , argv[2+*argi], 8))
		break;
	}
	if (j>=curr->nentries) {
	    fprintf(stderr, "Warning: couldn't find resource %s in WAD file %s\n", argv[2+*argi], argv[*argi+1]);
	    pnlen = 0;
	    pndata = malloc(1);
	} else {
	    curr->u.realwad.passthru[j] = 0;
	    pnlen = Itohl(curr->u.realwad.entries[j].length);
	    fseek(fp, (long) Itohl(curr->u.realwad.entries[j].offset), SEEK_SET);
		 pndata = malloc((size_t)pnlen);
		 fread(pndata, (size_t)pnlen, 1, fp);
	}

	for (i=*argi+3;
	     argv[i] && 0!=strcmp("-", argv[i]) && 0!=strcmp("+", argv[i]);
	     i++) {
	    struct texture_gob	*gob;
	    for (j=0; j<curr->nentries; j++) {
		if (0==strncasecmp(curr->u.realwad.entries[j].name , argv[i],8))
		    break;
	    }
	    if (j>=curr->nentries) {
		fprintf(stderr, "Warning: couldn't find resource %s in WAD file %s\n", argv[i], argv[*argi+1]);
		continue;
	    }

	    curr->u.realwad.passthru[j] = 0;

	    gob = extra_gob();
	    gob->tlen = Itohl(curr->u.realwad.entries[j].length);
	    fseek(fp, (long) Itohl(curr->u.realwad.entries[j].offset), SEEK_SET);
		 gob->tdata = malloc((size_t)gob->tlen);
		 fread(gob->tdata, (size_t)gob->tlen, 1, fp);

	    gob->pndata = pndata;
	    gob->pnlen = pnlen;
	}

	if (i==*argi+3) {
	    struct texture_gob	*gob;
	    /* no textures! */
	    gob = extra_gob();
	    gob->tlen = 0;
	    gob->tdata = 0;
	    gob->pndata = pndata;
	    gob->pnlen = pnlen;
	}

	if (!(argv[i] && 0==strcmp("+", argv[i]))) {
	    /* don't leak */
	    nwads--;
	    free(allwads[nwads].u.realwad.entries);
	    free(allwads[nwads].u.realwad.passthru);
	} 

	*argi=i;
    }

}

/* loads a single raw resource for later writing with the -owad option */

void slurp_resource(char *fname, char *resname)
{
    FILE	*fp;
    struct wadcache	*curr;
	 INT32	size = 4096;

    /* special case for lame DOS */
    if (0==strcasecmp(fname, "/dev/null")) {
	curr = new_wadcache();
	curr->nentries = 1;
	upcopy8(curr->u.oneres.name, resname);
	curr->u.oneres.data = 0;
	curr->u.oneres.len = 0;
	return;
    }

    if (!(fp = fopen_b(fname, 0, 1)))
	exit(1);

    curr = new_wadcache();

    curr->nentries = 0;
    upcopy8(curr->u.oneres.name, resname);
    
    curr->u.oneres.data = malloc(size);
    curr->u.oneres.len = 0;
    while (1) {
	INT32	n = size - curr->u.oneres.len;
	int	i;
	if (((int)n)!=(i=fread(curr->u.oneres.data+(int)curr->u.oneres.len, 1, (size_t)n, fp))) {
	    if (i<0) {
		fprintf(stderr, "error during read from file %s\n", fname);
	    } else if (i==0)
		break;
	}
	curr->u.oneres.len += i;
	if (size<=curr->u.oneres.len)
	    curr->u.oneres.data = realloc(curr->u.oneres.data, size*=2);
    }

	 curr->u.oneres.data = realloc(curr->u.oneres.data, (size_t)curr->u.oneres.len);

    fclose_b(fp);
}

/* writes a PWAD file containing the compiled texture and pnames
   resource.  It also has any raw resources and -iwads that were
   loaded with + instead of - */

void spurt_owad(char *fname, char *pname, char *tname)
{
    FILE	*fp;

	 if (!(fp = fopen_b(fname, 1, 1)))
	return;

    {
	struct wadhandle	*wh;
	char	*buffer;
	INT32	buffersize;
	int	i,j;

	buffer = malloc((size_t)buffersize=4096);

	wh = open_wadfile(fp);

	write_entry_to_wadfile(wh, output_texture, otlen, tname);
	write_entry_to_wadfile(wh, output_pnames, oplen, pname);
	for (i=nwads-1; i>=0; i--) {
	    if (allwads[i].nentries) {
		for (j=0; j<allwads[i].nentries; j++) {
			 INT32	len;
		    if (!allwads[i].u.realwad.passthru[j])
			continue;	/* skip this entry */
		    fseek(allwads[i].u.realwad.fp, (long)Itohl(allwads[i].u.realwad.entries[j].offset),
			  SEEK_SET);
			 len = Itohl(allwads[i].u.realwad.entries[j].length);
		    if (len>buffersize)
			buffer = realloc(buffer, (size_t)(buffersize = len));
			 if (len != 0 && 1!=fread(buffer, (size_t)len, 1, allwads[i].u.realwad.fp)) {
			fprintf(stderr, "error reading resource from wad during owad stage: ") ;
			perror("");
		    }
		    write_entry_to_wadfile(wh, buffer, len,
					   allwads[i].u.realwad.entries[j].name);
		}
	    } else {
		write_entry_to_wadfile
		    (wh, allwads[i].u.oneres.data,
		     allwads[i].u.oneres.len, allwads[i].u.oneres.name);
	    }
	}
	close_wadfile(wh);
    }

    fclose_b(fp);
}

/* read all the bytes from a file and return them.

   fname is the name of the file.  If it happens to be "-" then we
   read from stdin.

   len is a pointer to where we should store the length of the data.

   the pointer to the beginning of the data is returned from the function.

   */

void *slurp_file(char *fname, INT32 *len)
{
    int	n;
    char	*rawdata;
    int	rd_len, rd_size;

    FILE	*fp;

	 if (!(fp = fopen_b(fname, 0, 1)))
	return 0;

    rawdata = malloc(rd_size=4096);
    rd_len=0;

    while (1) {
	n = rd_size-rd_len;
	if (0 >= (n=fread(rawdata+rd_len, 1, n, fp))) {
	    if (n==0)
		break;
	    if (n<0) {
		fprintf(stderr, "Error reading texture resource from stdin: ");
		perror("");
		exit(1);
	    }
	}
	/* fprintf(stderr, "read %d\n", n); */
	rd_len += n;
	rawdata = realloc(rawdata, rd_size*=2);
    }

    /* data is now read into rawdata. */

    realloc(rawdata, rd_len);	/* we don't need all those other bytes
				   anymore */
    fclose_b(fp);

    *len = rd_len;
    return rawdata;
}

char ** load_argfile(char *argv0, char *fname, int *argc)
{
    FILE	*fp;
    char	buf[1024];
    char	**rval;
    int	size;

    *argc=0;

    if (!(fp = fopen_b(fname, 0, 0)))
	return 0;

    rval = (char**)malloc(sizeof(*rval)*(size=32));

    rval[(*argc)++] = argv0;

    while (0!=fgets(buf, sizeof(buf), fp)) {
	int	len = strlen(buf);
	if (buf[len-1]=='\n')
	    buf[--len] = 0;
	if (len==0)
	    continue;		/* totally empty lines don't count */
	if (*argc>=size)
	    rval = (char**)realloc(rval, sizeof(*rval)*(size*=2));
	rval[(*argc)++] = strdup(buf);
    }

    fclose_b(fp);

    return rval;
}

void Pause( char *message )
{
#ifdef __MSDOS__
    fprintf(stderr, "%s", message );
    getch();
#endif
}

void usage(char *progname)
{
    fprintf(stderr, "Usage:\n");
    fprintf(stderr, "  %s  options  \n", progname);
    fprintf(stderr, "\n");
    fprintf(stderr, "  Options:\n");
    fprintf(stderr, "-argfile file	specifies a file containing command-line arguments (one\n");
    fprintf(stderr, "		per line).  All command-line arguments following this one\n");
    fprintf(stderr, "		will be discarded\n");
    fprintf(stderr, "-bp file	specifies input binary pnames file (such as the PNAMES\n");
    fprintf(stderr, "		resource)\n");
    fprintf(stderr, "-bt file	specifies input binary texture file (such as the TEXTURE1\n");
    fprintf(stderr, "		or TEXTURE2 resource)\n");
    fprintf(stderr, "-ap file	specifies an ascii pnames file (1 name to a line)\n");
    fprintf(stderr, "-at file	specifies an ascii texture file (format NYD)\n");
    fprintf(stderr, "-res file name	loads a raw resource from file and gives it name.  This\n");
	 fprintf(stderr, "		option is only effective if the -owad option is used.\n");
    fprintf(stderr, "-iwad file +	specifies an input WAD file to be included in the WAD file\n");
    fprintf(stderr, "		written by the -owad option.  Good for merging in mission\n");
    fprintf(stderr, "		data.\n");
    fprintf(stderr, "-iwad file pname texture ... [+-]\n");
    fprintf(stderr, "		specifies an input WAD file.  The arguments after the\n");
    Pause( "Press any key to see the rest of help...\n" );
    fprintf(stderr, "		file name specify what to search for in the input wad\n");
    fprintf(stderr, "		file's directory to extract the pnames and texture data.\n");
    fprintf(stderr, "		A + or - is required and terminates the list of texture\n");
    fprintf(stderr, "		resources.  If it's a plus and the -owad option is used\n");
    fprintf(stderr, "		then all resources BESIDES the pname and textures are\n");
    fprintf(stderr, "		passed through unmodified to the output WAD.\n");
    fprintf(stderr, "-obp file	specifies an output binary pnames filename\n");
    fprintf(stderr, "-obt file	specifies an output binary texture filename\n");
    fprintf(stderr, "-oap file	specifies an output ascii pnames filename\n");
    fprintf(stderr, "-oat file	specifies an output ascii texture filename\n");
    fprintf(stderr, "-owad file pname texture\n");
    fprintf(stderr, "		specifies an output WAD file.  The arguments before the\n");
    fprintf(stderr, "		file name specify what to call the pnames and texture data\n");
    fprintf(stderr, "		in the output wad file's directory.\n\n");
}

/* this is where we store the output filenames. */
struct {
    char	*obp, *obt,
		*oap, *oat,
		*owad, *pname, *texture;
} outfiles = {
    0, 0,
    0, 0,
    0, 0, 0
};

#define DEMAND_ARGC(off)	do { if ((i+off)>=argc) { fprintf(stderr, "Too few arguments to option %s (expected %d)\n", argv[i], off); exit(1); } } while (0)

int main(int argc, char **argv)
{
    void	*currpnames=0;
    INT32	currpnlen;
    int	i;

    fprintf(stderr, "Textures utility version 1.04\n");
    fprintf(stderr, "Copyright (C) 1994 Robert Forsman\n");
	 fprintf(stderr, "Gnu General Public License\n");
    fprintf(stderr, "Ported to DOS by Robert Forsman and David Allen\n");

    /* parse arguments */

    if (argc<2) {
	usage(argv[0]);
	exit(1);
    }

    for (i=1; i<argc; i++) {
	if (0==strcmp(argv[i], "-argfile")) {
	    DEMAND_ARGC(1);
	    argv = load_argfile(argv[0], argv[i+1], &argc);
	    if (argv==0) {
		fprintf(stderr, "argfile load aborting\n");
		exit(1);
	    }
	    i=0;
	} else if (0==strcmp(argv[i], "-bp")) {
	    DEMAND_ARGC(1);
	    currpnames = slurp_file(argv[i+1], &currpnlen);
		 i++;
	} else if (0==strcmp(argv[i], "-bt")) {
		 struct texture_gob	*curr = extra_gob();
	    if (currpnames==0) {
		fprintf(stderr, "Must specify at least one pnames resource before specifying a binary\ntexture file.  Aborting.\n");
		exit(1);
	    }
		 DEMAND_ARGC(1);
		 curr->tdata = slurp_file(argv[i+1], &curr->tlen);
	    curr->pndata = currpnames;
	    curr->pnlen = currpnlen;
	    i++;
	} else if (0==strcmp(argv[i], "-ap")) {
	    DEMAND_ARGC(1);
		 currpnames = read_ascii_pnames(argv[i+1], &currpnlen);
	    i++;
	} else if (0==strcmp(argv[i], "-at")) {
		 struct texture_gob	*curr = extra_gob();
	    DEMAND_ARGC(1);
	    make_gob_from_textfile(argv[i+1], curr);
	    i++;
	} else if (0==strcmp(argv[i], "-iwad")) {
	    DEMAND_ARGC(3);
	    slurp_iwad(argv, &i);
	} else if (0==strcmp(argv[i], "-res")) {
	    DEMAND_ARGC(2);
	    slurp_resource(argv[i+1], argv[i+2]);
	    i+=2;
	} else if (0==strcmp(argv[i], "-obt")) {
	    DEMAND_ARGC(1);
	    if (outfiles.obt!=0) {
		fprintf(stderr, "One output-binary-texture file already specified, ignoring second\n");
	    } else {
		outfiles.obt = argv[i+1];
	    }
	    i++;
	} else if (0==strcmp(argv[i], "-obp")) {
	    DEMAND_ARGC(1);
	    if (outfiles.obp!=0) {
		fprintf(stderr, "One output-binary-pnames file already specified, ignoring second\n");
	    } else {
		outfiles.obp = argv[i+1];
	    }
	    i++;
	} else if (0==strcmp(argv[i], "-oat")) {
	    DEMAND_ARGC(1);
	    if (outfiles.oat!=0) {
		fprintf(stderr, "One output-ascii-texture file already specified, ignoring second\n");
	    } else {
		outfiles.oat = argv[i+1];
	    }
	    i++;
	} else if (0==strcmp(argv[i], "-oap")) {
	    DEMAND_ARGC(1);
	    if (outfiles.oap!=0) {
		fprintf(stderr, "One output-ascii-pnames file already specified, ignoring second\n");
	    } else {
		outfiles.oap = argv[i+1];
	    }
	    i++;
	} else if (0==strcmp(argv[i], "-owad")) {
	    DEMAND_ARGC(3);
	    if (outfiles.owad!=0) {
		fprintf(stderr, "One output WAD file already specified, ignoring second\n");
	    } else {
		outfiles.owad = argv[i+1];
		outfiles.pname = argv[i+2];
		outfiles.texture = argv[i+3];
	    }
	    i+=3;
	} else {
	    static int once = 0;
	    fprintf(stderr, "Unknown option %s\n", argv[i]);
	    if (!once) {
		usage(argv[0]);
		once = 1;
	    }
	}
    }

    /* merge all the texture resources we've been given into one big gob. */

    compile_textures();

    /* start writing output files. */

    if (outfiles.owad!=0) {
	spurt_owad(outfiles.owad, outfiles.pname, outfiles.texture);
    }
    if (outfiles.obp) {
	FILE	*fp = fopen_b(outfiles.obp, 1, 1);
	if (fp) {
	    fwrite(output_pnames, oplen, 1, fp);
	    fclose_b(fp);
	}
    }
    if (outfiles.obt) {
	FILE	*fp = fopen_b(outfiles.obt, 1, 1);
	if (fp) {
	    fwrite(output_texture, otlen, 1, fp);
	    fclose_b(fp);
	}
    }
    if (outfiles.oap) {
	FILE	*fp = fopen_b(outfiles.oap, 1, 0);
	if (fp) {
	    int	i,j;
	    for (i=0; i<output_pnames->count; i++) {
		for (j=0; output_pnames->names[i][j] && j<8; j++)
		    putc(output_pnames->names[i][j], fp);
		putc('\n', fp);
	    }
	    fclose_b(fp);
	}
    }
    if (outfiles.oat) {
	FILE	*fp = fopen_b(outfiles.oat, 1, 0);
	if (fp) {
	    pretty_print(fp);
	    fclose_b(fp);
	}
    }

    return 0;
}
