/*

    waddir.c: routines to read and write DOOM WAD files.
    Copyright (C) 1994  Robert H. Forsman Jr.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "wad.h"

struct wadhandle *open_wadfile(FILE *fp)
{
    struct wadhandle	*rval;

    rval = malloc(sizeof(*rval));

    rval->fp = fp;
    rval->basepos = ftell(fp);
    rval->nentries = 0;
	 rval->entries = malloc(sizeof(*rval->entries)*((int)rval->space = 16));
    rval->inmission=0;

    fseek(fp, (long)12, SEEK_CUR);

    return rval;
}

void scan_wadfile(struct wadhandle *wh)
{
    struct wadfile	header;
	 long	offset;

    fseek(wh->fp, (long)wh->basepos, SEEK_SET);
    fread(&header, sizeof(header), 1, wh->fp);
    wh->nentries = Itohl(header.nentries);
	 offset = Itohl(header.offset);

    fseek(wh->fp, (long)offset-sizeof(header), SEEK_CUR);

    if (wh->space < wh->nentries) {
	 wh->entries = realloc
		 (wh->entries, sizeof(*wh->entries)*(int)(wh->space = wh->nentries));
    }
	 fread(wh->entries, sizeof(*wh->entries), (size_t)wh->nentries, wh->fp);
}

char *mapres[MIS_COUNT] = {
    "THINGS",
    "LINEDEFS",
    "SIDEDEFS",
    "VERTEXES",
    "SEGS",
    "SSECTORS",
    "NODES",
    "SECTORS",
    "REJECT",
    "BLOCKMAP",
};

void *get_resource_from_wadfile(struct wadhandle *wh, INT32 *len, char *name)
{
    void	*rval;
    int	i;

    for (i=0; i<wh->nentries; i++) {
	if (0==strncasecmp(wh->entries[i].name, name, 8))
	    break;
    }
    if (i >= wh->nentries)
	return 0;
	 rval = malloc((size_t)(*len = Itohl(wh->entries[i].length)));
    fseek(wh->fp, wh->entries[i].offset + wh->basepos, SEEK_SET);
	 fread(rval, (size_t)*len, 1, wh->fp);
    return rval;
}

int write_entry_to_wadfile(struct wadhandle *wf,
				 void *data, INT32 len, char *name)
{
    struct dw_direntry	*curr;

    {
	int	i;
	for (i=0; i<MIS_COUNT; i++) {
	    if (0==strncasecmp(name, mapres[i], 8))
		break;
	}
	if (i<MIS_COUNT) {
	    if (!wf->inmission) {
		fprintf(stderr, "Warning:  Found mission resource %s outside ExMx context\n", name);
	    } else {
		if ((1<<i) & wf->bitvec) {
		    fprintf(stderr, "Warning:  Found duplicate mission resource %s\n", name);
		}
		wf->bitvec |= (1<<i);
	    }
	} else {
	    if (wf->inmission) {
		if (wf->bitvec != (1<<MIS_COUNT)-1) {
		    fprintf(stderr, "incomplete mission entry.  Missing:\n");
		    for (i=0; i<MIS_COUNT; i++)
			if (! (wf->bitvec & (1<<i)))
			    fprintf(stderr, "%s\n", mapres[i]);
		}
		wf->inmission = 0;
		wf->bitvec=0;
	    }
	    for (i=0; i<wf->nentries; i++) {
		if (0==strncasecmp(wf->entries[i].name, name, 8)) {
		    fprintf(stderr, "error, wad directory already has one resource named %s.  Not writing extra.\n", name);
		    return 0;
		}
	    }
	}

	if (2==sscanf(name, "E%dM%d", &i, &i)) {
	    wf->inmission=1;
	    wf->bitvec=0;
	}
    }

    if (wf->nentries>=wf->space) {
	wf->entries = realloc(wf->entries,
					sizeof(*wf->entries) * ((size_t) wf->space *=2));
    }

	 curr = &wf->entries[(int)(wf->nentries++)];

    curr->offset = htoIl(ftell(wf->fp) - wf->basepos);
    upcopy8(curr->name, name);
    curr->length = htoIl(len);

    fprintf(stderr,"wrote resource ");
    write8(name, stderr);
    fprintf(stderr," of len %ld at offset %ld\n",
	    Itohl(curr->length), Itohl(curr->offset));

    if (len!=0 && 1!=fwrite(data, (size_t)len, 1, wf->fp)) {
	fprintf(stderr, "error writing resource %s to wad file\n", name);
	return 0;
    }
    return 1;
}

int close_wadfile(struct wadhandle *wf)
{
    long	offset = ftell(wf->fp) - wf->basepos;
    struct wadfile	header;
    if (wf->nentries !=
	fwrite(wf->entries, sizeof(*wf->entries), (size_t)wf->nentries, wf->fp)) {
	fprintf(stderr, "Error while writing wad directory\n");
	perror("");
	return 0;
    }
    fseek(wf->fp, wf->basepos, SEEK_SET);
    memcpy(header.magic, "PWAD", 4);
    header.nentries = htoIl(wf->nentries);
    header.offset = htoIl(offset);
    if (1!= fwrite(&header, sizeof(header), 1, wf->fp)) {
	fprintf(stderr, "Error while writing wad header\n");
	perror("");
	return 0;
    }
    free(wf->entries);
    free(wf);
    return 1;
}
